package VISD::Apache2; # $Id: Apache2.pm 93 2020-09-25 13:47:07Z abalama $
use strict;
use warnings FATAL => 'all';
use utf8;

=encoding utf-8

=head1 NAME

VISD::Apache2 - mod_perl2 VISD interface

=head1 VERSION

Version 1.01

=head1 SYNOPSIS

    # Apache2 config section
    Alias /visd/ "/usr/share/visd/www/"
    <Directory "/usr/share/visd/www">
        Options +FollowSymlinks
        AllowOverride None
        DirectoryIndex index.html
        Require all granted
    </Directory>
    <Location /cetus-data>
       SetHandler modperl
       PerlResponseHandler VISD::Apache2
    </Location>

=head1 DESCRIPTION

To use the functionality of this module, you must edit the virtual-host section of
the Apache2 WEB server configuration file by specifying authorization data.

See C<INSTALL> file first!

=over 4

=item B<default>

    GET https://example.com/cetus-data

Returns main last_data.json file to client

    GET https://example.com/cetus-data?host=foo

Returns last_data.json filr of the foo host to client

=item B<graph>

    GET <BASE_URL>/cetus-data?action=graph&host=<HOST_NAME>
      &group=<GROUP_NAME>&name=<OBJECT_NAME>&zoom=<ZOOM>

This template shows how to get a graph. See L<GROUPS> and L<ZOOMS>

    GET https://example.com/cetus-data?action=graph

Returns 3h system graph of current host

    GET https://example.com/cetus-data?action=graph&zoom=1d&host=foo&group=net&name=eth0

Returns 1d network (iface=eth0) graph of foo host

=back

=head2 METHODS

=over 4

=item B<handler>

Handler for Apache2 WEB server

=back

=head2 GROUPS

Available graph groups

=over 4

=item B<sys>

System graph - shows CPU usage, RAM usage, Swap usage and IO delay

=item B<load>

System load graph - shows load average 1, 5 and 15

=item B<dev_usage>

Device (Disk) usage - this graph shows disk usage history

The device (name) required!

=item B<dev_io>

Device (Disk) IO - shows how many was read and write bytes

The device (name) required!

=item B<net>

Network interface usage - shows incoming anf outgoing traffic

The interface (name) required!

=back

=head2 ZOOMS

Supported zoom intervals

=over 4

=item B<3h>

Data for the 3 last hours

=item B<1d>

Data for the last day

=item B<1w>

Data for the last week

=item B<1m>

Data for the last month

=item B<1y>

Data for the last year

=back

=head1 HISTORY

See C<Changes> file

=head1 DEPENDENCIES

L<mod_perl2>

=head1 TO DO

See C<TODO> file

=head1 BUGS

* none noted

=head1 SEE ALSO

L<VISD>, L<mod_perl2>, C<INSTALL>

=head1 AUTHOR

Serż Minus (Sergey Lepenkov) L<http://www.serzik.com> E<lt>abalama@cpan.orgE<gt>

=head1 COPYRIGHT

Copyright (C) 1998-2020 D&D Corporation. All Rights Reserved

=head1 LICENSE

This program is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

See C<LICENSE> file and L<https://dev.perl.org/licenses/>

=cut

use vars qw/ $VERSION /;
$VERSION = "1.01";

use mod_perl2;
use Apache2::RequestIO ();
use Apache2::RequestRec ();
use Apache2::RequestUtil ();
use Apache2::Response ();
use Apache2::ServerRec ();
use Apache2::ServerUtil ();
use Apache2::Connection ();
use Apache2::Const -compile => qw/ :common :http /;
use Apache2::Log;
use Apache2::Util ();
use APR::Const -compile => qw/ :common /;
use APR::Table ();

use CGI -compile => qw/ :all /;
use CTK::Util qw/bload sharedstatedir/;
use CTK::Serializer;
use CTK::ConfGenUtil;
use CTK::TFVals qw/ :ALL /;
use File::Spec;

use VISD;
use VISD::Tools;

use constant {
        CONTENT_TYPE_LD     => "application/json",
        CONTENT_TYPE_PNG    => "image/png",
        CONTENT_TYPE_SVG    => "image/svg+xml",
    };

sub handler {
    my $r = shift;
    Apache2::RequestUtil->request($r);
    $r->handler('modperl');
    my $q = CGI->new();
    my $datadir = File::Spec->catdir(sharedstatedir(), VISD::PROJECTNAMEL);
    my $serializer = CTK::Serializer->new(VISD::SERIALIZER_FORMAT, attrs => VISD::SERIALIZER_ATTRS);

    # Get last data from file
    my $ldf = File::Spec->catfile($datadir, VISD::LAST_DATA_FILE);
    return _error(404 => sprintf("File \"%s\" not found", $ldf)) unless -e $ldf;
    #$r->log_error(sprintf("RRDcacheD socket: %s", RRDTOOLS_SOCKET)) if RRDTOOLS_SOCKET;
    my $data = {};
    if (my $bindata = bload($ldf, 1)) {
        $data = $serializer->deserialize($bindata);
        return _error(500 => $serializer->error) unless $serializer->status;
    } else {
        return _error(411 => sprintf("Can't load file \"%s\"", $ldf));
    }

    # Get ahosts from last data
    my $hostname = value($data, "hostname") // DEFAULT_HOSTNAME;
    my $ahosts = array($data, "ahosts");
    push @$ahosts, $hostname;
    my $agraphs = hash($data, "agraphs");
    my $net_interfaces = hash($data, "net/interfaces");
    my $fs_dev = hash($data, "fs/dev");
    $agraphs->{$hostname} = {
        interfaces => [keys(%$net_interfaces)],
        devices => [keys(%$fs_dev)],
    };

    # Get host
    my $host = $q->param("host") // $hostname;
    $host =~ s/[^a-z0-9\-_.]//ig;
    return _error(404 => "Host not found!") unless $host;
    return _error(400 => 'The "host" query string parameter incorrect!') unless $agraphs->{$host};

    #
    # Index: Returns last_data.json
    #
    if (!$q->param("action")) {
        $ldf = File::Spec->catfile($datadir, lc($host), VISD::LAST_DATA_FILE) if $host ne $hostname;
        my $len = (-s $ldf) || 0;
        return _error(411 => sprintf("File \"%s\" is empty", $ldf)) unless $len;
        $r->content_type(CONTENT_TYPE_LD);
        $r->set_content_length($len);
        return _error(507 => sprintf("Can't send file \"%s\" to output", $ldf)) if $r->sendfile($ldf) != APR::Const::SUCCESS;
        $r->rflush();
        return Apache2::Const::OK;
    }

    #
    # Graph
    #

    # Get group
    my $group = $q->param("group") // "sys";
    return _error_graph(400 => 'The "group" query string parameter incorrect!') unless grep {$_ eq $group} @{ (VISD::Tools::AVAILABLE_GROUPS()) };

    # Get zooms
    my $zoom = $q->param("zoom") // "3h";
    return _error_graph(400 => 'The "zoom" query string parameter incorrect!') unless grep {$_ eq $zoom} @{ (VISD::Tools::AVAILABLE_ZOOMS()) };

    # Get name
    my $name = $q->param("name") // "";
    if ($group eq 'net') {
        return _error_graph(400 => 'The "name" query string parameter incorrect!') unless grep {$_ eq $name} @{ (array($agraphs, $host, "interfaces")) };
    } elsif ($group =~ /^dev/) {
        return _error_graph(400 => 'The "name" query string parameter incorrect!') unless grep {$_ eq $name} @{ (array($agraphs, $host, "devices")) };
    }

    # Get RRD file name
    my $rrdf = File::Spec->catfile($datadir, lc($host), get_rrdfn($group, $name));

    # Plotting (trying)
    my ($rrdstatus, $rrdmsg);
    {
        local $@; # protect existing $@
        eval {
            ($rrdstatus, $rrdmsg) = rrd_graph(
                file => $rrdf,
                group => $group,
                zoom => $zoom,
            );
        };
        if ($@) {
            $rrdstatus = 0;
            $rrdmsg = $@;
        }
    }
    return _error_graph(500 => sprintf("Can't plot RRD graph: %s", $rrdmsg)) unless $rrdstatus;

    # OUTPUT
    $r->content_type(CONTENT_TYPE_PNG);
    $r->headers_out->set('Accept-Ranges', 'none');
    $r->set_content_length(length($rrdmsg) || 0);
    $r->print($rrdmsg);
    $r->rflush();
    return Apache2::Const::OK;
}

sub _error {
    my $err = pop;
    my $code = pop || Apache2::Const::SERVER_ERROR;

    my $r = Apache2::RequestUtil->request();
    $ENV{REDIRECT_ERROR_NOTES} = $err;
    $r->subprocess_env(REDIRECT_ERROR_NOTES => $err);
    $r->notes->set('error-notes' => $err);
    $r->log_error($err);
    my $msg = sprintf('{"status":0,"error":"%s"}', $err);

    $r->status($code);
    $r->content_type(CONTENT_TYPE_LD);
    $r->set_content_length(length($msg));
    $r->print($msg);
    $r->rflush();

    return $code;
}

sub _error_graph {
    my $err = pop;
    my $code = pop || Apache2::Const::SERVER_ERROR;

    my $r = Apache2::RequestUtil->request();
    $r->log_error($err);

    $r->content_type(CONTENT_TYPE_SVG);
    $r->headers_out->set('Accept-Ranges', 'none');

my $msg = <<SVG;
<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<svg xmlns="http://www.w3.org/2000/svg" id="error" version="1.1" width="640" height="260">
    <rect id="rect_error" width="100%" height="100%" style="fill: #CCCC88; fill-opacity:0.75" />
    <text id="text_title" x="25" y="50" style="font-weight: bold; font-size: 32; line-height:3.25; font-family: sans-serif; fill: #f24252">
      Error #$code
    </text>
    <text id="text_body" x="25" y="100" style="font-weight: normal; font-size: 18; line-height:1.25; font-family: sans-serif; fill: #f24252">
      $err
    </text>
</svg>
SVG

    $r->set_content_length(length($msg) || 0);
    $r->print($msg);
    $r->rflush();
    return Apache2::Const::OK;
}

1;

__END__
