package VISD::Tools; # $Id: Tools.pm 92 2020-09-24 19:14:42Z abalama $
use strict;
use utf8;

=encoding utf8

=head1 NAME

VISD::Tools - Various VISD tools

=head1 VERSION

Version 1.01

=head1 SYNOPSIS

    use VISD::Tools;

=head1 DESCRIPTION

This module contains various tools and utilities functions

=head1 VARIABLES

=head2 HAVE_RRDTOOLS

If this directive contains a true value, it means that the RRDtools package
is installed on your system and is ready to using

=head2 RRDTOOLS_SOCKET

    my $rrdcached = RRDTOOLS_SOCKET
       or die "Can't connect to rrdcached socket service";

Returns status of RRD service (rrdcached)

=head1 FUNCTIONS

=head2 dev2name

    my $name = dev2name("/dev/foo"); # returns "dev_foo"

Converts device paths to device names

=head2 duplicates

    my @result = duplicates(1,1,2,4,7,2,3,4,6,9); # returns 1,2,4

Returns a new list by stripping values in input list occuring
less than twice by comparing the values as hash keys

Author: Jens Rehsack; See L<List::MoreUtils>

=head2 fetch

    my ($status, $message) = fetch($url, $dir, $exp);

This function returns data from remote resource by specified URL.
The message contents error or document

=head2 get_pngfn

    my $file = get_pngfn($host, $group, $name, $zoom);
    my $file = get_rrdfn("localhost", "net" "eth0", "3h"); # Returns localhost_net_eth0_3h.png
    my $file = get_rrdfn("localhost", "dev_io" "/dev/null", "3h"); # Returns localhost_dev_io_dev_null_3h.png

Returns filename of PNG file by host, group, name and zoom

=head2 get_rrdfn

    my $file = get_rrdfn($group, $name);
    my $file = get_rrdfn("sys", "myname"); # Returns sys.rrd
    my $file = get_rrdfn("net", "myname"); # Returns net_myname.rrd
    my $file = get_rrdfn("dev_io", "/dev/null"); # Returns fs_dev_null.rrd

Returns filename of rrd file by group and name

=head2 req_or_avail

    my @sel = req_or_avail([qw/foo bar/], [qw/foo bar baz/]); # Returns foo, bar
    my @sel = req_or_avail([qw/foo bar quux/], [qw/foo bar baz/]); # Returns foo, bar
    my @sel = req_or_avail([qw/one two/], [qw/foo bar baz/]); # Nothing returns
    my @sel = req_or_avail([], [qw/foo bar baz/]); # Returns foo, bar, baz

=head2 rrd_create

    my ($rrdstatus, $rrdmsg) = rrd_create(
        file => "sys.rrd",
        type => "sys",
        step => 20,
    );

Creates RRD database for system data

=head2 rrd_error

    print rrd_error();

Returns RRDs error string

=head2 rrd_graph

    my ($rrdstatus, $rrdmsg) = rrd_graph(
        file => "sys.rrd",
        group => "sys",
        zoom => $zoom,
    );

Returns plotted graph data

=head2 rrd_update

    my ($rrdstatus, $rrdmsg) = rrd_update(
        file => "sys.rrd",
        type => "sys",
        ## data:
        cpu => 12,
        mem => 24,
        swp => 6,
        iod => 10,
    );

Update RRD database for system data

=head1 NON AUTO IMPORTED FUNCTIONS

=head2 set_fetch_timeout

    VISD::Tools::set_fetch_timeout(5);

Sets timeout for fetching

=head1 SEE ALSO

L<VISD>

=head1 AUTHOR

Serż Minus (Sergey Lepenkov) L<http://www.serzik.com> E<lt>abalama@cpan.orgE<gt>

=head1 COPYRIGHT

Copyright (C) 1998-2020 D&D Corporation. All Rights Reserved

=head1 LICENSE

This program is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

See C<LICENSE> file and L<https://dev.perl.org/licenses/>

=cut

use vars qw/$VERSION @EXPORT @EXPORT_OK/;
$VERSION = '1.01';

@EXPORT = (qw/
        HAVE_RRDTOOLS
        RRDTOOLS_SOCKET

        duplicates
        req_or_avail
        dev2name
        get_rrdfn
        get_pngfn
        fetch

        rrd_error
        rrd_create
        rrd_update
        rrd_graph
    /);
@EXPORT_OK = (@EXPORT, qw/
        RRDTOOLS_SOCKET_UNIX
        RRDTOOLS_SOCKET_INET
        RRDTOOLS_SOCKET_TIMEOUT

        AVAILABLE_ZOOMS
        AVAILABLE_TYPES
        AVAILABLE_GROUPS
        GROUP2TYPE_MAP

        set_fetch_timeout
    /);

use base qw/Exporter/;

use VISD;

use Carp qw/carp croak/;
use File::Spec;
use IO::Socket::INET;
use URI;
use Time::Local qw/timelocal/;
use LWP::Simple qw/$ua head mirror/;
use HTTP::Date qw/time2str/;
use CTK::Util qw/bload/;

use constant {
        FALSE                   => 0,
        TRUE                    => 1,
        FETCH_TIMEOUT           => 5,
        RRDTOOLS_SOCKET_UNIX    => "unix:/var/run/rrdcached.sock",
        RRDTOOLS_SOCKET_INET    => "127.0.0.1:42217",
        RRDTOOLS_SOCKET_TIMEOUT => 5, # 5 sec
        RRDTOOLS_RRA            => [ # RRA:CF:xff:steps:rows
            "RRA:AVERAGE:0.5:1:600",
            "RRA:AVERAGE:0.5:6:700",
            "RRA:AVERAGE:0.5:24:775",
            "RRA:AVERAGE:0.5:288:797",
            "RRA:MIN:0.5:1:600",
            "RRA:MIN:0.5:6:700",
            "RRA:MIN:0.5:24:775",
            "RRA:MIN:0.5:288:797",
            "RRA:MAX:0.5:1:600",
            "RRA:MAX:0.5:6:700",
            "RRA:MAX:0.5:24:775",
            "RRA:MAX:0.5:288:797",
        ],
        AVAILABLE_ZOOMS         => [qw/ 3h 1d 1w 1m 1y /],
        AVAILABLE_TYPES         => [qw/ sys net dev /],
        AVAILABLE_GROUPS        => [qw/ sys load net dev_io dev_usage /],
        GROUP2TYPE_MAP          => {
            sys         => "sys",
            load        => "sys",
            net         => "net",
            dev_io      => "dev",
            dev_usage   => "dev",
        },
        GRAPH_WIDTH     => 640,
        GRAPH_HEIGHT    => 260,
    };

# RRDs loading
my $_have_rrdtools = FALSE;
foreach my $idir (@INC) {
    if (-e File::Spec->catfile($idir, "RRDs.pm")) {
        $_have_rrdtools = TRUE;
        require RRDs;
        last;
    }
}
*HAVE_RRDTOOLS = sub () { return $_have_rrdtools };

# RRDs socket
my $_rrdtools_socket;
if (RRDTOOLS_SOCKET_UNIX =~ /^unix\:(.+)$/ and -S $1) {
    $_rrdtools_socket = RRDTOOLS_SOCKET_UNIX;
} elsif (_check_inet_socket(RRDTOOLS_SOCKET_INET)) {
    $_rrdtools_socket = RRDTOOLS_SOCKET_INET;
}
*RRDTOOLS_SOCKET = sub () { return $_rrdtools_socket };

sub _check_inet_socket {
    my $addr = shift || RRDTOOLS_SOCKET_INET;
    my $sock = IO::Socket::INET->new(
        PeerAddr    => $addr,
        Proto       => 'tcp',
        Timeout     => RRDTOOLS_SOCKET_TIMEOUT,
    ) or return 0; # Could not create socket INET: $!
    return 0 unless $sock->connected();
    $sock->close();
    return 1;
}

##
## RRD functions
##

sub rrd_error {
    return "No RRDs installed" unless HAVE_RRDTOOLS();
    return sprintf("RRDs error: %s", RRDs::error() // "unknown error");
}

# https://oss.oetiker.ch/rrdtool/doc/rrdcreate.en.html
sub rrd_create {
    my %opts = @_;
    return (FALSE, "No RRDs installed") unless HAVE_RRDTOOLS();
    my $file = $opts{file};
    croak "No file specified!" unless $file;
    return (TRUE, "") if -e $file;  # Skip: The file already exists!
    my $type = $opts{type}; # sys | load | net | dsk | dskio
    my $step = $opts{step} || 300;
    my $heartbeat = $step * 2;

    # Set general args
    my @args = (
        "--step" => $step,
    );

    # Add socket to args
    push @args, "--daemon" => RRDTOOLS_SOCKET()
        if RRDTOOLS_SOCKET() && RRDs->VERSION >= 1.5;

    if ($type eq 'sys') {
        push @args, # DS:ds-name:DST:heartbeat:min:max
            sprintf("DS:cpu:GAUGE:%d:0:100", $heartbeat),
            sprintf("DS:mem:GAUGE:%d:0:100", $heartbeat),
            sprintf("DS:swp:GAUGE:%d:0:100", $heartbeat),
            sprintf("DS:iod:GAUGE:%d:0:100", $heartbeat),
            sprintf("DS:avg1:GAUGE:%d:0:U", $heartbeat),
            sprintf("DS:avg5:GAUGE:%d:0:U", $heartbeat),
            sprintf("DS:avg15:GAUGE:%d:0:U", $heartbeat);
    } elsif ($type eq 'net') {
        push @args, # DS:ds-name:DST:heartbeat:min:max
            sprintf("DS:in:COUNTER:%d:U:U",, $heartbeat),
            sprintf("DS:out:COUNTER:%d:U:U", $heartbeat);
    } elsif ($type eq 'dev') {
        push @args, # DS:ds-name:DST:heartbeat:min:max
            sprintf("DS:usage:GAUGE:%d:0:100", $heartbeat),
            sprintf("DS:read:COUNTER:%d:U:U",, $heartbeat),
            sprintf("DS:write:COUNTER:%d:U:U", $heartbeat);
    } else {
        return (FALSE, "Unrecognized type!");
    }
    push @args, @{ (RRDTOOLS_RRA) };

    # Call!
    RRDs::create($file, @args);

    # Result
    return (FALSE, rrd_error()) if RRDs::error();
    return (TRUE, "");
}

# https://oss.oetiker.ch/rrdtool/doc/rrdupdate.en.html
sub rrd_update {
    my %opts = @_;
    return (FALSE, "No RRDs installed") unless HAVE_RRDTOOLS();
    my $file = $opts{file};
    croak "No file specified!" unless $file;
    my $type = $opts{type}; # sys | load | net | dsk | dskio
    my $timestamp = $opts{timestamp} // "N";

    # Set general args
    my @args = ();

    # Add socket to args
    push @args, "--daemon" => RRDTOOLS_SOCKET() if RRDTOOLS_SOCKET();

    my $template;
    if ($type eq 'sys') {
        $template = "cpu:mem:swp:iod:avg1:avg5:avg15";
    } elsif ($type eq 'net') {
        $template = "in:out";
    } elsif ($type eq 'dev') {
        $template = "usage:read:write";
    } else {
        return (FALSE, "Unrecognized type!");
    }
    if ($template) {
        push @args, "--template" => $template unless RRDTOOLS_SOCKET();
        my @tmp = ();
        foreach my $t (split(/\:/, $template)) {
            push @tmp, $opts{$t} // "U";
        }
        push @args, sprintf("%s:%s", $timestamp, join(":", @tmp)) if @tmp;
    }

    # Call!
    RRDs::update($file, @args);

    # Result
    return (FALSE, rrd_error()) if RRDs::error();
    return (TRUE, "");
}

# https://oss.oetiker.ch/rrdtool/doc/rrdgraph.en.html
sub rrd_graph {
    my %opts = @_;
    return (FALSE, "No RRDs installed") unless HAVE_RRDTOOLS();
    my $file = $opts{file};
    croak "No file specified!" unless $file;

    # Set general args
    my @args = (
        "-a" => "PNG", # --imgformat
        "-b" => 1000, # --base
        "-D", # --full-size-mode
        "-w" => GRAPH_WIDTH, # --width
        "-h" => GRAPH_HEIGHT, # --height
        "--border" => 1,
        "-c" => "ARROW#222244", # --color
        "-E", # --slope-mode
        "-e" => "now", # --end
        "-W" => time2str(),
    );

    # Add socket to args
    push @args, "--daemon" => RRDTOOLS_SOCKET() if RRDTOOLS_SOCKET();

    # Set zoom
    my $title_sfx = "unknown";
    my $vborder;
    $opts{zoom} //= AVAILABLE_ZOOMS()->[0]; # 3h
    if ($opts{zoom} eq '3h') {
        $title_sfx = "the last 3 hours (5 min average)";
        push @args, "-s" => "-3h"; # --start
        push @args, "-x" => "MINUTE:15:HOUR:1:HOUR:1:0:%H:%M"; # --x-grid
    } elsif ($opts{zoom} eq '1d') {
        $title_sfx = "the last day (5 min average)";
        push @args, "-s" => "-24h"; # --start
        push @args, "-x" => "MINUTE:30:HOUR:1:HOUR:3:0:%a %H:%M"; # --x-grid
        $vborder = sprintf("VRULE:%s#EE0000", _vborder());
    } elsif ($opts{zoom} eq '1w') {
        $title_sfx = "the last week (30 min average)";
        push @args, "-s" => "-7d"; # --start
        push @args, "-x" => "HOUR:6:DAY:1:DAY:1:86400:%a %d/%m"; # --x-grid
        $vborder = sprintf("VRULE:%s#EE0000", _vborder("weekly"));
    } elsif ($opts{zoom} eq '1m') {
        $title_sfx = "the last month (2 hour average)";
        push @args, "-s" => "-1mon"; # --start
        push @args, "-x" => "DAY:1:DAY:7:DAY:7:0:%d/%m"; # --x-grid
        push @args, "-S" => 3600*2; # --step
        $vborder = sprintf("VRULE:%s#EE0000", _vborder("monthly"));
    } elsif ($opts{zoom} eq '1y') {
        $title_sfx = "the last year (1 day average)";
        push @args, "-s" => "-12mon"; # --start
        push @args, "-x" => "MONTH:1:MONTH:3:MONTH:1:2592000:%b"; # --x-grid
        push @args, "-S" => 3600*24; # --step
        $vborder = sprintf("VRULE:%s#EE0000", _vborder("yearly"));
    }

    # Set titles and labels
    $opts{group} //= AVAILABLE_GROUPS()->[0];
    if ($opts{group} eq 'sys') { # sys
        push @args, "-t" => sprintf("System in %s", $title_sfx); # --title
        push @args, "-v" => "Usage, %"; # --vertical-label
        # Data
        push @args,
            sprintf("DEF:cpu=%s:cpu:AVERAGE", $file),
            sprintf("DEF:mem=%s:mem:AVERAGE", $file),
            sprintf("DEF:swp=%s:swp:AVERAGE", $file),
            sprintf("DEF:iod=%s:iod:AVERAGE", $file),
            "VDEF:cpu_cur=cpu,LAST",
            "VDEF:cpu_avg=cpu,AVERAGE",
            "VDEF:cpu_min=cpu,MINIMUM",
            "VDEF:cpu_max=cpu,MAXIMUM",
            "VDEF:mem_cur=mem,LAST",
            "VDEF:mem_avg=mem,AVERAGE",
            "VDEF:mem_min=mem,MINIMUM",
            "VDEF:mem_max=mem,MAXIMUM",
            "VDEF:swp_cur=swp,LAST",
            "VDEF:swp_avg=swp,AVERAGE",
            "VDEF:swp_min=swp,MINIMUM",
            "VDEF:swp_max=swp,MAXIMUM",
            "VDEF:iod_cur=iod,LAST",
            "VDEF:iod_avg=iod,AVERAGE",
            "VDEF:iod_min=iod,MINIMUM",
            "VDEF:iod_max=iod,MAXIMUM",
            "COMMENT:\\s",
            "COMMENT: ",
            "LINE1:cpu#00FF00:CPU usage   ",
            "GPRINT:cpu_cur:[Cur\\:%4.0lf%s",
            "GPRINT:cpu_avg:Avg\\:%4.0lf%s",
            "GPRINT:cpu_min:Min\\:%4.0lf%s",
            "GPRINT:cpu_max:Max\\:%4.0lf%s]\\l",
            "COMMENT: ",
            "LINE1:mem#0000FF:Memory usage",
            "GPRINT:mem_cur:[Cur\\:%4.0lf%s",
            "GPRINT:mem_avg:Avg\\:%4.0lf%s",
            "GPRINT:mem_min:Min\\:%4.0lf%s",
            "GPRINT:mem_max:Max\\:%4.0lf%s]\\l",
            "COMMENT: ",
            "LINE2:swp#FFFF00:Swap usage  ",
            "GPRINT:swp_cur:[Cur\\:%4.0lf%s",
            "GPRINT:swp_avg:Avg\\:%4.0lf%s",
            "GPRINT:swp_min:Min\\:%4.0lf%s",
            "GPRINT:swp_max:Max\\:%4.0lf%s]\\l",
            "COMMENT: ",
            "LINE1:iod#FF0000:IO delay    ",
            "GPRINT:iod_cur:[Cur\\:%4.0lf%s",
            "GPRINT:iod_avg:Avg\\:%4.0lf%s",
            "GPRINT:iod_min:Min\\:%4.0lf%s",
            "GPRINT:iod_max:Max\\:%4.0lf%s]\\l",
            "COMMENT:\\s";
    } elsif ($opts{group} eq 'load') { # load
        push @args, "-t" => sprintf("Load average in %s", $title_sfx); # --title
        push @args, "-v" => "Units"; # --vertical-label
        push @args, "-o", "--units=si"; # --logarithmic
        # Data
        push @args,
            sprintf("DEF:avg1=%s:avg1:AVERAGE", $file),
            sprintf("DEF:avg5=%s:avg5:AVERAGE", $file),
            sprintf("DEF:avg15=%s:avg15:AVERAGE", $file),
            "VDEF:avg1_cur=avg1,LAST",
            "VDEF:avg1_avg=avg1,AVERAGE",
            "VDEF:avg1_min=avg1,MINIMUM",
            "VDEF:avg1_max=avg1,MAXIMUM",
            "VDEF:avg5_cur=avg5,LAST",
            "VDEF:avg5_avg=avg5,AVERAGE",
            "VDEF:avg5_min=avg5,MINIMUM",
            "VDEF:avg5_max=avg5,MAXIMUM",
            "VDEF:avg15_cur=avg15,LAST",
            "VDEF:avg15_avg=avg15,AVERAGE",
            "VDEF:avg15_min=avg15,MINIMUM",
            "VDEF:avg15_max=avg15,MAXIMUM",
            "COMMENT:\\s",
            "COMMENT: ",
            "LINE1:avg1#339933:Load average 1 ",
            "GPRINT:avg1_cur:[Cur\\:%4.0lf%s",
            "GPRINT:avg1_avg:Avg\\:%4.0lf%s",
            "GPRINT:avg1_min:Min\\:%4.0lf%s",
            "GPRINT:avg1_max:Max\\:%4.0lf%s]\\l",
            "COMMENT: ",
            "LINE1:avg5#003333:Load average 5 ",
            "GPRINT:avg5_cur:[Cur\\:%4.0lf%s",
            "GPRINT:avg5_avg:Avg\\:%4.0lf%s",
            "GPRINT:avg5_min:Min\\:%4.0lf%s",
            "GPRINT:avg5_max:Max\\:%4.0lf%s]\\l",
            "COMMENT: ",
            "LINE1:avg15#333300:Load average 15",
            "GPRINT:avg15_cur:[Cur\\:%4.0lf%s",
            "GPRINT:avg15_avg:Avg\\:%4.0lf%s",
            "GPRINT:avg15_min:Min\\:%4.0lf%s",
            "GPRINT:avg15_max:Max\\:%4.0lf%s]\\l",
            "COMMENT:\\s",
    } elsif ($opts{group} eq 'net') { # net
        push @args, "-t" => sprintf("Net traffic in %s", $title_sfx); # --title
        push @args, "-v" => "Bits per Second"; # --vertical-label
        push @args, "-o", "--units=si"; # --logarithmic
        # Data
        push @args,
            sprintf("DEF:inoctets=%s:in:AVERAGE", $file),
            sprintf("DEF:outoctets=%s:out:AVERAGE", $file),
            "CDEF:in=inoctets,8,*",
            "CDEF:out=outoctets,8,*",
            "VDEF:in_cur=in,LAST",
            "VDEF:in_avg=in,AVERAGE",
            "VDEF:in_min=in,MINIMUM",
            "VDEF:in_max=in,MAXIMUM",
            "VDEF:out_cur=out,LAST",
            "VDEF:out_avg=out,AVERAGE",
            "VDEF:out_min=out,MINIMUM",
            "VDEF:out_max=out,MAXIMUM",
            "COMMENT:\\s",
            "COMMENT: ",
            "AREA:in#00FF00:Incoming",
            "GPRINT:in_cur:[Cur\\:%4.0lf%s",
            "GPRINT:in_avg:Avg\\:%4.0lf%s",
            "GPRINT:in_min:Min\\:%4.0lf%s",
            "GPRINT:in_max:Max\\:%4.0lf%s]\\l",
            "LINE1:in#00DD00",
            "COMMENT: ",
            "LINE1:out#0000FF:Outgoing",
            "GPRINT:out_cur:[Cur\\:%4.0lf%s",
            "GPRINT:out_avg:Avg\\:%4.0lf%s",
            "GPRINT:out_min:Min\\:%4.0lf%s",
            "GPRINT:out_max:Max\\:%4.0lf%s]\\l",
            "COMMENT:\\s";
    } elsif ($opts{group} eq 'dev_io') { # dev_io
        push @args, "-t" => sprintf("Disk IO in %s", $title_sfx); # --title
        push @args, "-v" => "Bytes per Second"; # --vertical-label
        push @args, "-o", "--units=si"; # --logarithmic
        # Data
        push @args,
            sprintf("DEF:read=%s:read:AVERAGE", $file),
            sprintf("DEF:write=%s:write:AVERAGE", $file),
            "VDEF:read_cur=read,LAST",
            "VDEF:read_avg=read,AVERAGE",
            "VDEF:read_min=read,MINIMUM",
            "VDEF:read_max=read,MAXIMUM",
            "VDEF:write_cur=write,LAST",
            "VDEF:write_avg=write,AVERAGE",
            "VDEF:write_min=write,MINIMUM",
            "VDEF:write_max=write,MAXIMUM",
            "COMMENT:\\s",
            "COMMENT: ",
            "LINE1:read#66FF00:Bytes read   ",
            "GPRINT:read_cur:[Cur\\:%4.0lf%s",
            "GPRINT:read_avg:Avg\\:%4.0lf%s",
            "GPRINT:read_min:Min\\:%4.0lf%s",
            "GPRINT:read_max:Max\\:%4.0lf%s]\\l",
            "COMMENT: ",
            "LINE1:write#FF6600:Bytes written",
            "GPRINT:write_cur:[Cur\\:%4.0lf%s",
            "GPRINT:write_avg:Avg\\:%4.0lf%s",
            "GPRINT:write_min:Min\\:%4.0lf%s",
            "GPRINT:write_max:Max\\:%4.0lf%s]\\l",
            "COMMENT:\\s";
    } elsif ($opts{group} eq 'dev_usage') { # dev_usage
        push @args, "-t" => sprintf("Disk usage in %s", $title_sfx); # --title
        push @args, "-v" => "Usage, %"; # --vertical-label
        # Data
        push @args,
            sprintf("DEF:usage=%s:usage:AVERAGE", $file),
            "VDEF:usage_cur=usage,LAST",
            "VDEF:usage_avg=usage,AVERAGE",
            "VDEF:usage_min=usage,MINIMUM",
            "VDEF:usage_max=usage,MAXIMUM",
            "COMMENT:\\s",
            "COMMENT: ",
            "AREA:usage#00CC66:Usage",
            "GPRINT:usage_cur:[Cur\\:%4.0lf%s",
            "GPRINT:usage_avg:Avg\\:%4.0lf%s",
            "GPRINT:usage_min:Min\\:%4.0lf%s",
            "GPRINT:usage_max:Max\\:%4.0lf%s]\\l",
            "LINE1:usage#009933",
            "COMMENT:\\s";
    }
    push @args, $vborder if defined $vborder;

    # Call!
    my $res = RRDs::graphv('-', @args); # we do not really store data into the file

    # Result
    return (FALSE, rrd_error()) if RRDs::error();
    return (TRUE, $res->{image});
}

##
## General functions
##

sub duplicates { # See List::MoreUtils 4+
    my %seen = ();
    my $k;
    my $seen_undef;
    grep { 1 < (defined $_ ? $seen{$k = $_} : $seen_undef) }
      grep { defined $_ ? not $seen{$k = $_}++ : not $seen_undef++ } @_;
}

sub req_or_avail {
    my $req = shift;
    my $avl = shift;
    my @av = $avl && (ref($avl) eq 'ARRAY') ? (@$avl) : ();
    return (@av) unless $req && (ref($req) eq 'ARRAY') && scalar(@$req);
    return (duplicates(@$req, @av));
}

sub dev2name {
    my $dev = shift;
    return "" unless defined $dev;
    my @splt = grep {length $_} split(/[\/\\]+/, $dev);
    return join "_", @splt;
}

sub get_rrdfn($$) {
    my ($group, $name) = @_;
    my $type = GROUP2TYPE_MAP()->{$group} || "sys";
    if ($type eq "net") {
        return sprintf("net_%s.rrd", $name);
    } elsif ($type eq "dev") {
        return sprintf("fs_%s.rrd", dev2name($name));
    }
    return sprintf("sys.rrd");
}
sub get_pngfn($$$$) {
    my ($host, $group, $name, $zoom) = @_;
    my $type = GROUP2TYPE_MAP()->{$group} || "sys";
    if ($type eq "net") {
        return sprintf("%s_%s_%s_%s.png", lc($host), $group, $name, $zoom);
    } elsif ($type eq "dev") {
        return sprintf("%s_%s_%s_%s.png", lc($host), $group, dev2name($name), $zoom);
    }
    return sprintf("%s_%s_%s.png", lc($host), $group, $zoom);
}

sub set_fetch_timeout {
    my $to = shift || FETCH_TIMEOUT;
    $ua->timeout($to);
}

sub fetch($;$$) {
    my $url = shift;
    my $dir = shift || File::Spec->curdir();
    my $exp = shift || 0;
    my $file = File::Spec->catfile($dir, VISD::LAST_DATA_FILE);
    my $curtime = time();
    my $mtime = 0; # 0 - file not exists
    if (-e $file) {
        $mtime = (stat($file))[9] || 0;
    }

    # URL Check
    unless (defined $url) {
        carp("File to fetching is not specified");
        return (0, "File to fetching is not specified");
    }

    # Working with remote file
    if (!$mtime || ($mtime+$exp) < $curtime) { # if NO file OR it is too old
        # Create URI instance
        my $uri = URI->new($url);

        # HEAD request
        my ($content_type, $document_length, $modified_time, $expires, $server) = head($uri);
        $modified_time ||= $curtime;
        #print(sprintf("HEAD Response:"
        #    ."\n\tContent-type:\t%s"
        #    ."\n\tContent-length:\t%s"
        #    ."\n\tModified:\t%s"
        #    ."\n\tServer:\t\t%s\n",
        #        defined($content_type) ? $content_type : '',
        #        defined($document_length) ? $document_length : 0,
        #        defined($modified_time) ? scalar(localtime($modified_time)) : '',
        #        defined($server) ? $server : ''
        #    ));

        # Check response data
        return (0, "Remote document not exists or is empty") unless $document_length;

        #printf("Local\t: %s\nRemote\t: %s\n",
        #    defined($mtime) ? scalar(localtime($mtime)) : 'none',
        #    defined($modified_time) ? scalar(localtime($modified_time)) : 'none',
        #);

        # Remote file is newer than local
        if (!$mtime || ($mtime < $modified_time)) { # Remote document is newer!
            # Delete file
            if ($mtime) { # realy it is: if -e $file
                return(0, sprintf("Can't delete file %s: %s",$file, $!)) unless unlink($file);
            }

            # Download!
            #print("Downloading!\n");
            my $code = mirror($uri, $file);
            unless (($code >= 200) && ($code < 400)) { # errorr
                return(0, sprintf("%d Can't download %s to %s: %s", $code, $url, $file, $!));
            }
        }
    }

    # Loading
    unless ((-e $file) && -s $file) { # File not exists
        return(0, sprintf("Can't download %s to %s. File not exist", $url, $file));
    }

    # Ok. Loading!
    if (my $bindata = bload($file, 1)) {
        return(1, $bindata);
    }
    return(0, sprintf("Can't loading file %s", $file));
}

sub _vborder {
    my $g = shift || '';

    my $off = 0; # daily
    if ($g eq "yearly") {
        my $offday = (localtime(time()))[7] || 0; # year day: 0..364
        $off = (-1 * 3600 * 24 * $offday);
    } elsif ($g eq "monthly") {
        my $offday = (localtime(time()))[3] || 0; # month day: 1..28,29,30,31
        $off = (-1 * 3600 * 24 * ($offday-1));
    } elsif ($g eq "weekly") {
        my $offday = (localtime(time()))[6] || 0; # week day: 0..6
           $offday = 7 if $offday == 0;
        $off = (-1 * 3600 * 24 * ($offday-1));
    }

    my ($sec,$min,$hour,$mday,$mon,$year) = localtime(time() + $off);
    return timelocal(0,0,0,$mday,$mon,$year);
}

1;

__END__
